from supermutes.utils import get_new_obj, get_class_registrar


CLASS_REGISTER = {}

dotify = lambda obj: get_new_obj(CLASS_REGISTER, obj)


def register(old_class, new_class):
    CLASS_REGISTER[old_class] = new_class


def reset_mapping():
    register(dict, DotDict)
    register(list, DotList)


class DotList(list):
    """
    A list that allows dot notation access to its items.

    For example:

    >> d = DotList(['a', 'b', 'c'])
    >> d._0
    'a'
    >> d._1
    'b'
    >> d._2
    'c'
    >> d._3
    IndexError

    """
    __metaclass__ = get_class_registrar("DotList", list, register)

    def __init__(self, *args, **kwargs):
        list.__init__(self, *args, **kwargs)
        for i, value in enumerate(self):
            self[i] = value

    def __getattr__(self, attr):
        try:
            index = int(attr.strip('_'))
            return self[index]
        except ValueError:
            pass
        return list.__getattr__(self, attr)

    def __setattr__(self, attr, value):
        try:
            index = int(attr.strip('_'))
            self[index] = value
        except ValueError:
            pass
        list.__setattr__(self, attr, dotify(value))

    def __setitem__(self, index, value):
        list.__setitem__(self, index, dotify(value))


class DotDict(dict):
    """
    A dictionary that allows dot notation access to its values.

    For example:

    >>  d = DotDict({
            'a': 12,
            'b': {
                'something': 5
            }
        })
    >>  d.a
    12
    >>  d.b.something
    5
    >>  d.c = 99
    >>  del d['a']
    >>  del d['b']
    >>  d
    {'c': 99}
    """
    __metaclass__ = get_class_registrar("DotDict", dict, register)

    def __init__(self, *args, **kwargs):
        dict.__init__(self, *args, **kwargs)
        for key, value in self.items():
            self[key] = value

    def __setitem__(self, key, value):
        dict.__setitem__(self, key, dotify(value))

    __getattr__ = dict.__getitem__
    __setattr__ = __setitem__


reset_mapping()



def get_new_obj(register, obj):
    for old_class, new_class in register.items():
        if isinstance(obj, old_class) and not isinstance(obj, new_class):
            return new_class(obj)
    return obj


def get_class_registrar(super_name, map_to_class, registration_func):
    class RegisterClass(type):
        """
        A metaclass used for registering any subclass as the main converter

        If this is used as a metaclass, then upon declaring a class that
        inherits from a class named ``super_name``, ``registration_func`` will
        be called with ``map_to_class`` and the cls currently being declared.
        """
        def __init__(cls, name, bases, attrs):
            try:
                base_names = [c.__name__ for c in bases]
                if super_name in base_names:
                    registration_func(map_to_class, cls)
            except NameError:
                return
    return RegisterClass



"A collection of readonly immutables"
from supermutes.utils import get_new_obj, get_class_registrar

CLASS_REGISTER = {}


def register(old_class, new_class):
    CLASS_REGISTER[old_class] = new_class


def reset_mapping():
    register(dict, ReadOnlyDict)
    register(list, ReadOnlyList)


class ReadOnlyClassException(Exception):
    pass

readonly = lambda obj: get_new_obj(CLASS_REGISTER, obj)


class ReadOnlyBaseClass():

    def method_not_allowed(self, *args, **kwargs):
        raise ReadOnlyClassException("Cannot write to object.")

    __setitem__ = __setattr__ = append = insert = method_not_allowed


class ReadOnlyList(ReadOnlyBaseClass, list):

    __metaclass__ = get_class_registrar("ReadOnlyList", list, register)

    def __getitem__(self, index):
        return readonly(list.__getitem__(self, index))


class ReadOnlyDict(ReadOnlyBaseClass, dict):

    __metaclass__ = get_class_registrar("ReadOnlyDict", dict, register)

    def __getitem__(self, key):
        return readonly(dict.__getitem__(self, key))


reset_mapping()
